<?php

namespace Drupal\cloud\Traits;

use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Entity\EntityInterface;

/**
 * The trait checking permission.
 */
trait AccessCheckTrait {

  /**
   * Checks if the user can access the cloud config entity with owner check.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity object.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   Run access checks for this account.
   * @param string $owner_permission
   *   The permission for the user who is owner.
   * @param string $not_owner_permission
   *   The permission for the user who is not owner.
   *
   * @return \Drupal\Core\Access\AccessResult
   *   Allowed or forbidden, neutral if tempstore is empty.
   */
  protected function allowedIfCanAccessCloudConfigWithOwner(
    EntityInterface $entity,
    AccountInterface $account,
    $owner_permission,
    $not_owner_permission
  ) {
    if (!$account->hasPermission('view all cloud config entities')
      && !$account->hasPermission('view ' . $entity->getCloudContext())) {
      return AccessResult::neutral();
    }

    if ($account->hasPermission($not_owner_permission)) {
      return AccessResult::allowed();
    }
    else {
      if ($account == NULL || $entity->getOwner() == NULL) {
        return AccessResult::neutral();
      }

      return AccessResult::allowedIf(
        $account->hasPermission($owner_permission) &&
        ($account->id() == $entity->getOwner()->id())
      );
    }
  }

  /**
   * Checks if the user can access the cloud config entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity object.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   Run access checks for this account.
   * @param string $permission
   *   The permission for the user.
   *
   * @return \Drupal\Core\Access\AccessResult
   *   Allowed or forbidden, neutral if tempstore is empty.
   */
  protected function allowedIfCanAccessCloudConfig(
    EntityInterface $entity,
    AccountInterface $account,
    $permission
  ) {
    if (!$account->hasPermission('view all cloud config entities')
      && !$account->hasPermission('view ' . $entity->getCloudContext())) {
      return AccessResult::neutral();
    }

    return AccessResult::allowedIfHasPermission($account, $permission);
  }

}
